from cms.include import *
from flask import abort, Response
import uuid
import traceback
import simplejson as json


class BaseViewMeta(type):
    def __new__(mcl, name, bases, nmspc):
        new_class = super(BaseViewMeta, mcl).__new__(mcl, name, bases, nmspc)

        module_path = nmspc['__module__']

        # TODO custom options and endpoint

        if not nmspc.get('__url__', None):
            return new_class

        if module_path.startswith('modules') or module_path.startswith('custom'):
            module_name = module_path.split('.')[-3]
            if module_name != 'index':
                url = '/' + module_name + nmspc['__url__']
            else:
                url = nmspc['__url__']
            new_class.real_url = url
            endpoint = module_name + '.' + name

            new_class.jinja_env = mcl._create_jinja_env(mcl, module_name)

            APP.add_url_rule(url, endpoint, lambda *args, **kwargs: new_class().dispatch(*args, **kwargs), methods=['GET', 'POST', 'PUT'])

            new_class.cms_action_list = []
            for action_name, action_method in new_class.__dict__.items():
                if hasattr(action_method, "cms_is_action"):
                    new_class.cms_action_list.append(action_name.upper())
        else:
            raise Exception('Вьюхи можем определять только в modules и переопределять их в custom')

        return new_class

    def _create_jinja_env(cls, module_name):
        options = {
            'extensions' : ['jinja2.ext.autoescape', 'jinja2.ext.with_'],
            'autoescape' : APP.select_jinja_autoescape,
        }
        # if 'auto_reload' not in options:
        #     if self.config['TEMPLATES_AUTO_RELOAD'] is not None:
        #         options['auto_reload'] = self.config['TEMPLATES_AUTO_RELOAD']
        #     else:
        #         options['auto_reload'] = self.debug
        jinja_loader = jinja2.FileSystemLoader([os.path.join(PROJECT_DIR, dir) for dir
                                                                                     in ['custom/{}/templates'.format(module_name),
                                                                                         'modules/{}/templates'.format(module_name),
                                                                                         'common/templates',
                                                                                         'cms/templates',
                                                                                         './']])
        options['loader'] = jinja_loader
        rv = jinja2.Environment(**options)
        from flask.helpers import url_for, get_flashed_messages
        from flask.globals import request, session, g

        rv.globals.update(APP.jinja_env.globals)
        rv.globals.update(
            url_for=url_for,
            get_flashed_messages=get_flashed_messages,
            config=config,
            request=request,
            session=session,
            g=g,
            csrf_token=lambda: flask.Markup(
                '<input name="csrf_token" type="hidden" value={} />'.format(cms.views.get_csrf_token())
            ) if session.get('csrf_token') else ''
        )

        rv.filters = APP.jinja_env.filters
        return rv


class BaseView(object, metaclass=BaseViewMeta):
    def dispatch(self, *args, **kwargs):
        method = request.environ['REQUEST_METHOD'].upper()
        if '_method' in request.values:
            method = request.values['_method'].upper()
        if 'method' in request.values:
            method = request.values['method'].upper()
        if 'action' in request.values:
            method = request.values['action'].upper()
        if method in self.cms_action_list:
            return getattr(self, method.lower())(*args, **kwargs)
        abort(400)

    def render_with_params(self, tempalte_name, **context):
        template = self.jinja_env.get_template(tempalte_name)
        return template.render(**context)

    def render(self, tempalte_name):
        """!!! Метод пробрасывает неймспейс view в шаблон"""
        # for frame_info in inspect.stack(context=0):
        #     print(frame_info)
        #     # print(dir(frame_info.frame.f_code))
        #     print(frame_info.frame.f_locals, frame_info.frame.f_trace)
        namespace_params = {}
        # Сходу не понятно как по коду найти саму функцию, у которой есть признак view: f.cms_is_action == True
        # Пока просто считаем, что нас вызывает view напрямую: stack[0] - мы, stack[1] - view
        view_frame = inspect.stack(context=0)[1].frame
        namespace_params.update(view_frame.f_globals)
        namespace_params.update(view_frame.f_locals)

        def rename_key(old, new):
            if old in namespace_params:
                namespace_params[new] = namespace_params[old]
                del namespace_params[old]

        # Fix namespace
        rename_key('self', 'view_self')
        rename_key('args', 'view_args')
        rename_key('kwargs', 'view_kwargs')
        return self.render_with_params(tempalte_name, **namespace_params)

    # Так тоже можно, но трэш
    # @staticmethod
    # def render_with_namespace_script(tempalte_name):
    #     """return eval(self.render_with_namespace_script(tempalte_name)) """
    #     return f'''self.jinja_env.get_template('{tempalte_name}').render(l=locals(), g=globals())'''


def action(f):
    f.cms_is_action = True
    return f


def methods(*methods):
    def real_action(f):
        def wrapper(*args, **kwargs):
            if methods and request.method not in methods:
                abort(405)
            return f(*args, **kwargs)

        return wrapper

    return real_action


### REST::::

class RestDocView(BaseView):
    '''
        Класс для вывода документации по всем rest-вьюхам в проекте. Его нужно зарегистрировать
    '''
    @classmethod
    def register_rest_view(cls, view_class):
        if not getattr(cls, 'rest_view_class_list', None):
            cls.rest_view_class_list = []
        if view_class not in cls.rest_view_class_list:
            cls.rest_view_class_list.append(view_class)

    def dispatch(self):
        return self.render_doc()

    def render_doc(self):
        return self.render('rest_doc.html', class_list=RestDocView.rest_view_class_list)


def rest_action(*param_list):
    def rest_action_wrap(f):
        rest_f = RestViewFunc(f=f, param_list=param_list)
        return rest_f

    return rest_action_wrap


class RestViewFuncParam(object):
    def __init__(self, name, help='', cast_func=None, required=True):
        self.name = name
        self.help = help
        self.cast_func = cast_func
        # TODO: required=False
        self.required = required

    def __str__(self):
        s = 'Параметр {}: {}'.format(self.name, self.help)
        if self.cast_func:
            s += '. Кастуется функцией {}'.format(self.cast_func)
        return s


class RestViewFunc(object):
    def __init__(self, f, param_list):
        self.cms_is_action = True
        self.f = f
        self.param_list = param_list
        self.name = f.__name__
        if not getattr(f, '__doc__', None):
            logging.warning('Напишите строку с документацией для rest-функции {}!'.format(f))
            self.doc = ''
        else:
            self.doc = f.__doc__

    def __get__(self, instance, owner):
        if instance is None:
            return self
        d = self
        mfactory = lambda self, *args, **kw: d(self, *args, **kw)
        mfactory.__name__ = self.f.__name__
        return mfactory.__get__(instance, owner)

    def __call__(self, instance):
        f_args = {}
        for param in self.param_list:
            if param.name not in request.values:
                return instance.render('rest_error.html', msg='Забыли указать параметр {}'.format(param.name)), 400
            param_value = request.values[param.name]

            if param.cast_func:
                try:
                    param_value = param.cast_func(param_value)
                except ValueError:
                    return instance.render('rest_error.html', msg='Параметр {}={} не проходит каст {}'.format(
                        param.name, param_value, param.cast_func)), 400

            f_args[param.name] = param_value

        try:
            f_res = self.f(instance, **f_args)
        except Exception as e:
            uuid_str = str(uuid.uuid4())
            logging.error('REST error {}({}) <{}> {}'.format(self.f, f_args, uuid_str, e))
            logging.error('{}'.format(traceback.format_exc()))
            APP.cms_exception_mail(e)
            return instance.render('rest_error.html', msg='REST запрос завершился с ошибкой. Сообщите код ошибки администратору: {}'.format(uuid_str)), 500
        return Response(json.dumps(f_res, use_decimal=True, ensure_ascii=False).encode('utf-8'), mimetype="application/json; charset=utf-8")


class RestViewMeta(BaseViewMeta):
    def __new__(mcl, name, bases, nmspc):
        new_class = super(RestViewMeta, mcl).__new__(mcl, name, bases, nmspc)
        RestDocView.register_rest_view(new_class)
        return new_class


class RestView(BaseView, metaclass=RestViewMeta):
    def rest_f_iterator(self):
        return [f for f in self.__dict__.values() if isinstance(f, RestViewFunc)]
